%Aqual2 -- second cut at analyzing ADNI phantom data
%usage:  Aqual2(filename,optVerboseLevel);
%Analyze ADNI phantom from filename
%first pass: only nifti1 formatted data will be accepted
%verbose is optional:  0 is quiet (except for fatal errors), 1 reports any time a fallback
%procedure is hit and >=2 enables maximal reporting
%return values <0 should come with the help message, 0 no detected errors
%1 input data error, 2 sphere finding error
%function status=Aqual2(filename,verbose)
function [status,results,phan]=Aqual2(filename,writeLabels,destinationDirectory,verbose)
status=0;
results=[];
if ~exist('filename','var')
    help 'Aqual2';
    status=-1 ;
    return;
end

f=strtrim(filename);
if ~strcmp(f,filename)
    warning('filename had leading and/or trailing whitespace trimmed');
end
filename=f;
clear f;

Aqual_version=2.0;
started=datestr(now);
if ~exist('verbose','var')
    verbose=0;
end

if ~exist('destinationDirectory','var')
    destinationDirectory=[];
end

if ~exist('writeLabels','var')
    writeLabels=0;
end

if(verbose>1)
    disp('loading data');
end

pattern = '(.*)\.gz$';
isCompressed=0;
ocell=regexp(filename,pattern,'tokens');
if length(ocell)>0
    % then input is gzipped ;
    isCompressed=1;
    % ok we kludge things  -- make a tempdir
    myTemp=tempname;
    if(verbose>1)
        disp('uncompressing data into temporary directory');
        myTemp
    end
    realFile=gunzip(filename,myTemp);
    realFile=realFile{1};
    if(verbose>1)
        disp('file created:');
        disp(realFile);
    end
else
    realFile=filename;
end

[ret,phan]=Aqual2_loadNifti(realFile,verbose);
if ret~=0
    status=10 ;
    disp('failed loading data');
    return;
end
if(verbose>1)
    disp('done loading data');
end

if isCompressed>0
    disp('done reading temporary copy -- removing temporary file and temporary directory');
    delete(realFile);
    try
        rmdir(myTemp);
    catch
        try
            disp('temp directory was not empty -- trying with s option');
            rmdir(myTemp,'s');
        catch
            estring=sprintf('temp directory was not removed -- %s\n',myTemp);
            warning(estring);
        end
    end
end
% now Otsu thresholding will fail if data is not at least float so we check
if ~isfloat(phan.img)
    phan.img=single(phan.img)
end


phan.filename=filename;
if ~isempty(destinationDirectory)
    [dir,base,ext]=fileparts(phan.fileprefix);
    clear dir;
    phan.outputFilePrefix=fullfile(destinationDirectory,[base,ext]);
else
    if isCompressed>0
        pattern = '(.*)\.nii\.gz$';
        ocell=regexp(filename,pattern,'tokens');
        phan.outputFilePrefix=ocell{1}{1};
    else
        phan.outputFilePrefix=phan.fileprefix;
    end
end
% a little bit of setup work  -- this is the map against which we will
% guide the searches
% support multiple location sets
locations(1)=SpherePositions();
locations(2)=SpherePositionsAlt();
phanKeep=phan;
for theoryCounter=1:length(locations)
   phan=phanKeep;
    theory=locations(theoryCounter);
    phan.minSphereSeparation=theory.minSphereSeparation* 0.75;
    phan.voxelVolume=prod(phan.hdr.dime.pixdim(2:4));
    phan.masterLabelImg=zeros(size(phan.img));

    disp('searching for 6cm sphere');
    [ret,phan]=Aqual2_findSNRSphere(phan,theory,verbose);
    if ret~=0
        status=ret;
        return;
    end
    disp('done searching for 6m sphere');
    % the SNR sphere is not quite in the middle of the phantom  -- center is
    % off by  by 1.3mm in one dimension
    disp('searching for pole spheres');
    [ret,phan]=Aqual2_findSpherePoleSpheres(phan,theory,verbose);
    if ret~=0
        status=ret;
        return;
    end
    disp('done searching for pole spheres');

    disp('migrating theoretical sphere positions into local image space');
    % first off figure out the COM coordinates of the three spheres we have so
    % far -- since spheres are uniquely identified we don't have to hunt for
    % anything
    SO=phan.RASPositions(1,:);
    SY=phan.RASPositions(2,:);
    SX=phan.RASPositions(3,:);
    RAStoCOMTForm=getCOMTForm(SO,SX,SY);
    VOXtoRASTForm=maketform('affine',phan.RAS');
    COMtoRASTForm=fliptform(RAStoCOMTForm);
    RAStoVoxTForm=fliptform(VOXtoRASTForm);
    COMtoVoxTForm=maketform('composite',RAStoVoxTForm,COMtoRASTForm);
    phan.voxTheoryPositions=tformfwd([theory.Spheres.CanonicalCOM]',COMtoVoxTForm);
    phan.voxUnfittedTheoryPositions=phan.voxTheoryPositions;

    % now we perform a sanity check -- is what's the distance between the
    % phan.voxCentroids and the phan.voxTheoryPositions?  -- this should be
    % tiny
    th= phan.voxTheoryPositions(1:3,:);
    ph=phan.voxCentroids;
    d=th-ph;
    maxdiff=max(abs(d(:)));
    if maxdiff>10
        estring=sprintf('problems getting theory positions into image voxel coordinate system');
        Aqual2_writeFailureLog(phan.outputFilePrefix,estring);
        status=99; % this should never happen!
        return;
    end

    % ok now we want to work out a search strategy for which spheres to find
    % first.  This probably only matters because findSphereSmallSpheres
    % interates on the sphere locations and if you have spheres that are, for
    % example, all in a single plane then the updated sphere positions may do
    % mirror image flipping which makes the search space for sphere off the
    % plane incorrect.  findSphereSmallSpheres will ignore such tranformations
    % if they happen, but if we define a search which is more spatially
    % isotropic than the default ordering from spherePositions.m then it
    % should not happen-- we keep the check and ignore anyway.
    th= [theory.Spheres.CanonicalCOM]';
    %th= phan.voxTheoryPositions; use the theory. elements since we want a mm
    %shell, not measuring in voxels which in general will would be elliptical.

    center=theory.SNRSphereIndex;
    % thr=theory radius from center
    thr=[th(:,1)-th(center,1),th(:,2)-th(center,2),th(:,3)-th(center,3)];
    thr=thr.^2;
    thr=sqrt(sum(thr,2));
    thr(theory.CNRSphereIndex)=10*max(thr); % fake the CNR sphere to have a large radius so they
    % always get found last
    thr(1:3)=10*max(thr); % and ignore the SNR and pole spheres
    thr=abs(thr-60);  % start looking for spheres about 60 mm from center or SNR sphere
    % and walk in and out from that
    [thr,phan.searchOrder]=sort(thr);
    clear thr;
    phan.searchOrder=phan.searchOrder(1:158);


    %save;
    disp('we now have a coordinate system for spheres in the image');
    [ret,phan]=Aqual2_findSphereSmallSpheres(phan,verbose);

    if isequal(ret,50)
        disp('NOTE  ---- Bounding Box problems --- this may indicate that SNR sphere has come off its mount');
        disp('Try a different sphere location list');
        continue;
        % continue to try a different set of theory locations
    else
        if  ret~=0
            %undifferentiated error -- bail out
            status=ret;
            return;
        else
            % success -- this set of locations was good enough
            break;
        end
    end
end
disp('done searching for small spheres; now getting contrast sphere properties');
[ret,phan]=Aqual2_findSphereContrastSpheres(phan,theory,verbose);
if ret~=0
    status=ret;
    return;
end

disp('done getting contrast sphere properties');

% make diagnostic pictures
Aqual2_makePlots(phan,theory,verbose);


% now numerical analysis -- some of this is redundant with work
% done inside makePlots but we'll live with that in the name of
% compartmentalization


% get the RAS coordinates for the theory positions and observed
% positions; do OPP to get close and then figure out the scaling
%save;
goodSphereIndices=find(phan.foundSphere>0);
obsSpherePositions=phan.RASPositions(goodSphereIndices,:);
thSpherePositions=[theory.Spheres.CenterRAS]';
thSpherePositions=thSpherePositions(goodSphereIndices,:);
OPP=OrthogonalProcrustes(obsSpherePositions',thSpherePositions');
if det(OPP)<0
    disp(['warning: in Aqual2_makePlots Orthogonal Procustes introduced' ...
        ' a flip?']);
end
OPPTForm=maketform('affine',OPP');
theoryLocs=tformfwd(thSpherePositions,OPPTForm);
obs=obsSpherePositions';
th=theoryLocs';
D=obs'-th'; % nb the transpose is so that std(D) evaluates over spheres not directions
obsReg=double([obs;ones(1,size(obs,2))]);
thReg=double([th; ones(1,size(th,2))]);
T=thReg*obsReg'*inv(obsReg*obsReg');
T=T';
T(:,4)=[0 0 0 1]';
affineXF=maketform('affine',T);
% affineXF tell us how to scale rotated/translated theory positions
% in order to make them fit the observed positions.  In the limit
% that this tranform commutes with the OPP then affineXF should be
% diagonal -- certainly it's safe to assume that the angles are
% small and we don't have to do too many tricky things to get the
% signs right when we now deconstruct the affineXF

sintheta=T(3,2);
costheta=sqrt(1-sintheta*sintheta);
sinphi=T(3,1)/costheta;
cosphi=sqrt(1-sinphi*sinphi);
sinpsi=-1.0 *T(1,2)/costheta;
cospsi=sqrt(1-sinpsi*sinpsi);
sfac=[T(1,1)/(cospsi*cosphi +sinphi*sinpsi*sintheta),...
    T(2,2)/(costheta*cospsi),...
    T(3,3)/(costheta*cosphi)];
%disp(sprintf('estimated scale factors: %d, %d, %d',sfac))
% now start writing the text output file
txtfile=sprintf('%s_aqual2.txt',phan.outputFilePrefix);
fid=fopen(txtfile,'w');
if fid<0
    disp(sprintf('WARNING cannot open output file %s for writing', ...
        txtfile));
end
fprintf(fid,'Aqual2 output for: %s\n',phan.filename);
fprintf(fid,'this is Aqual version %g\n',Aqual_version);
fprintf(fid,'Executed %s by user %s on host %s \n',started, ...
    getenv('USER'),getenv('HOSTNAME'));
fprintf(fid,'Found %d fiducial spheres (should be 160)\n',length(goodSphereIndices));
missingSpheres=1-phan.foundSphere;
missingSpheres(theory.SNRSphereIndex)=0;
missingSpheres(theory.CNRSphereIndex)=0;
missingIndices = find(missingSpheres>0);
for i=1:length(missingIndices)
    fprintf(fid,'sphere %d not found\n',missingIndices(i));
end

fprintf(fid, 'Contrast Properties:\n');
fprintf(fid, 'S,N,SNR:%f,%f,%f\n',phan.S,phan.N,phan.SNR);
Corder=sort(theory.CNRSphereIndex); 
% nominally  the theory indices should be correctly ordered but 
% we'll sort them into ascending order just in case the list is differently
% ordered
Contrast=phan.intensityMean(Corder);
ContrastSD=phan.intensityStd(Corder);
Water=phan.waterPeak(Corder);
WaterSig=phan.waterWidth(Corder);

Plastic=phan.plasticPeak(Corder);
PlasticSig=phan.plasticWidth(Corder);
% we used to sort these by contrast intensity (commented out below)
% but for some field strengths the intensities are not well separated
% so now we use spatial location to get a unique support. 
%[Contrast,Corder]=sort(Contrast);
%Water=Water(Corder(:))
%WaterSig=WaterSig(Corder(:))
%Plastic=Plastic(Corder(:))
%PlasticSig=PlasticSig(Corder(:))

fprintf(fid, 'ContrastSphereMeanIntensities:%f;%f;%f;%f\n',Contrast(1),Contrast(2),Contrast(3),Contrast(4));
fprintf(fid, 'ContrastSphereIntensitySD:%f,%f,%f,%f\n', ContrastSD(1),ContrastSD(2),ContrastSD(3),ContrastSD(4));   
fprintf(fid,'ConstrastBackgroundWaterIntensities:%f;%f;%f;%f\n',Water(1),Water(2),Water(3),Water(4));
fprintf(fid,'ConstrastBackgroundWaterSD:%f;%f;%f;%f\n',WaterSig(1),WaterSig(2),WaterSig(3),WaterSig(4));
fprintf(fid,'ConstrastBackgroundPlasticIntensities:%f;%f;%f;%f\n',Plastic(1),Plastic(2),Plastic(3),Plastic(4));
fprintf(fid,'ConstrastBackgroundPlasticSD:%f;%f;%f;%f\n',PlasticSig(1),PlasticSig(2),PlasticSig(3),PlasticSig(4));

clear Contrast;
clear ContrastSD;
clear Water;
clear WaterSig;
clear Plastic;
clear PlasticSig;
fprintf(fid,'mean and sd of fiducial sphere average intensities:%f,%f\n',...
    mean(phan.intensityMean(goodSphereIndices)),...
    std(phan.intensityMean(goodSphereIndices)));

fprintf(fid, 'Summary Geometry Properties (values are R/L, A/P, S/I):\n');
fprintf(fid, 'Standard Deviation of Position Residuals (mm): %f,%f,%f\n',std(D));
fprintf(fid, ['Estimated Scale factors to linearly correct gradient' ...
    ' (unitless):%f,%f,%f\n'],sfac);
%fprintf(fid, ['unit vector scale factors' ...
%	      ' (unitless):%f,%f,%f\n'],sunit);
%pfac=(sfac-1.0)*100.0;

%fprintf(fid,'Correction Values translated for Philips Console (X,Y,Z) %f,%f,%f\n',pfac(2),pfac(1),pfac(3));
fprintf(fid,'Non-linear Evaluation:\n');
% loop over polynomial fit orders  -- this section liberated from original GE
% code
for order=1:5
    % BIG RED FLAG WARNING: THIS IS POSSIBLY NUMERICALLY UNSTABLE -- if you
    % don't promote to double Matlab may find rank deficient results when
    % doing right divide least squares fits.
    X=double(obsSpherePositions);
    Y=double(thSpherePositions);
    Coefficients = FitSpheres ( Y, X, order );
    Displacement = EvalPolyFit ( Coefficients, X, order );
    NewPoints = Y - Displacement;
    RMS = sqrt ( mean ( sum ( ( NewPoints' - X' ).^2 ) ) );
    disp ( sprintf ( 'Polynomial (%d order) RMS: %f', order, RMS ) );
    fprintf(fid,' Polynomial (%d order) RMS: %f\n',order, RMS );
end

thSpherePositions=[theory.Spheres.CenterRAS]';
% we want all theory  sphere positions  -- not just the ones we have

compositeXF=maketform('composite',affineXF,OPPTForm);
theoryLocs=tformfwd(thSpherePositions,compositeXF);

fprintf(fid,'Per sphere information table (can be read as a CSV file) :\n');
fprintf(fid,['sphere number,found,grey scale location x,grey scale location y,grey scale location z,centroid x,centroid y,'...
    'centroid z,cross corr x,cross corr  y,cross corr z,linear theory x,linear theory y,linear theory z,intensity average, background median,intensity ratio\n']) ;

for i=2:161
    bogusPosition=[10000 10000 10000];
    if phan.foundSphere(i)>0
        str= sprintf('%d,%d,%f,%f,%f',i,phan.foundSphere(i),phan.RASPositions(i,:));
        str= sprintf('%s,%f,%f,%f,%f,%f,%f',str,phan.RASCentroids(i,:),phan.RAScorrPositions(i,:));
        str = sprintf('%s,%f,%f,%f,%f,%f,%f\n',str,theoryLocs(i,:),phan.intensityMean(i),phan.backgroundIntensityMedian(i),phan.intensityMean(i)./phan.backgroundIntensityMedian(i));
    else
        str= sprintf('%d,%d,%f,%f,%f',i,phan.foundSphere(i),bogusPosition);
        str= sprintf('%s,%f,%f,%f,%f,%f,%f',str,bogusPosition,bogusPosition);
        str = sprintf('%s,%f,%f,%f,%f,%f,%f\n',str,theoryLocs(i,:),0,0,0);
    end
    fprintf(fid,str);
    if verbose>5
        disp(str);
    end
end
fclose(fid);

if writeLabels>0
    if(verbose>1)
        disp('writing segmentation labels');
    end
    p=phan;
    p.img=p.masterLabelImg;
    outFile=sprintf('%s_aqual2.nii',phan.outputFilePrefix);
    save_nii(p,outFile);
    if isCompressed>0
        if(verbose>1)
            disp('compressing segmentation labels');
        end
        gzip(outFile);
        delete(outFile);
    end
end
status=ret;
return;

