function Aqual(varargin)
% entry point for image analysis related to ADNI site qualification
%
% takes up to two arguments -- input directory containing the dicom
% images you are trying to segment and analyze and an output
% filename stub for the CSV output, greyscale image and segmented
% object map
%
% J Gunter  -- first cut 20050613
%
% NIFTI support added 20060223
%
%
try
  Aqual_version=1.3;
inDir=[];
outFile=['aqual.phrep'];
ofile='';
formatDICOM=[];
formatAnalyze=[];
formatNifti=[];
formatAVW=[];
if(length(varargin)==2)
    ofile=char(varargin(2));
    [outputdir,outputfile]=fileparts(ofile);
    inDir=char(varargin(1));
elseif(length(varargin)==1)
    inDir=char(varargin(1));
end
%  if isempty(ofile)
pattern='(\S*)\*.dcm';
try
    ocell=regexp(inDir,pattern,'tokens');
    ocell{1};
    fabOfile=sprintf('%s',char(ocell{1}));
    %disp(sprintf('fabOfile is %s\n',fabOfile));
    formatDICOM=1;
catch
    pattern='(\S*MR\S*)\*';
    try
        ocell=regexp(inDir,pattern,'tokens');
        ocell{1};
        fabOfile=sprintf('%s',char(ocell{1}));
        formatDICOM=1;
    catch
        % user seems not to have given us input with wildcard
        % check to see if it's a single img$ or hdr$
        pattern='(\S*)\.hdr';
        try
            ocell=regexp(inDir,pattern,'tokens');
            ocell{1};
            fabOfile=sprintf('%s',char(ocell{1}));
            % disp(sprintf('fabOfile is %s\n',fabOfile));
            formatAnalyze=1;
        catch
            pattern='(\S*)\.img';
            try
                ocell=regexp(inDir,pattern,'tokens');
                ocell{1};
                fabOfile=sprintf('%s',char(ocell{1}));
                %disp(sprintf('fabOfile is %s\n',fabOfile));
                formatAnalyze=1;
            catch
                pattern='(\S*)\.nii';
                try
                    ocell=regexp(inDir,pattern,'tokens');
                    ocell{1};
                    fabOfile=sprintf('%s',char(ocell{1}));
                    %disp(sprintf('fabOfile is %s\n',fabOfile));
                    formatNifti=1;
                catch
                    pattern='(\S*)\.avw';
                    try
                        ocell=regexp(inDir,pattern,'tokens');
                        ocell{1};
                        fabOfile=sprintf('%s',char(ocell{1}));
                        % disp(sprintf('fabOfile is %s\n',fabOfile));
                        formatAVW=1;
                    catch
                        % unless user gave us input throw a warning
                        if isempty(ofile)
                            warning(['cannot fabricate an output file name from your' ...
                                ' input']);
                        end
                    end
                end
            end
        end
    end
end
%save ofile;
if isempty(ofile)
    if isempty(fabOfile)
        warning(['no fabricated out name and no user supplied name?' ...
            ]);
    else
        ofile=fabOfile;
        clear fabOfile;
    end
end


% reassure user that something is happening

if ~isempty(formatDICOM)

    disp('loading DICOM series ... this may take awhile.')
    [Image, Headers] = LoadDICOMSeries ( inDir,1);
    disp('done loading DICOM series');
else

    if ~isempty(formatAVW)
        disp('loading AnalyzeAVW formatted data');
        Im=LoadAVW(inDir);
        Headers=Im;
        Image=Im.Image;
        clear Im;
    else
        if ~isempty(formatAnalyze)
            disp('loading Analyze 7.5 formatted data');
            Im=LoadAnalyzeImage(inDir);
            Headers=Im;
            Image=Im.Image;
            clear Im;
        else
            if ~isempty(formatNifti)
                disp('loading NIFTI formatted data');
                Im=LoadNiftiImage(inDir);
                Image=Im.Image;
                Headers=Im;
                clear Im;
            else
                warning('no input specified');
            end
        end
    end
end
try
    disp('segmenting spheres from image data');
    [Spheres LI] = SegmentSpheresADNI ( Headers, Image, []);
    disp('done segmenting');
    disp('checking for extra spheres');

    if length(Spheres.Spheres) > 165
        disp('trying to remove extra spheres');
        [Spheres, LI] = RemoveFalsePositives ( Spheres, SpherePositions, LI );
    end

    % now get SNR measurement
    % we need a mask of the SNR sphere
    %
    disp('getting SNR measurement');
    Mask= (LI==Spheres.SNRSphereIndex);
    [SNR,S,N]=CalculateSNRFromSphere ( Image, Mask);


    disp('setting up sphere locations as engineered');
    theory=SpherePositions();

    disp(['re-ordering spheres to get 1:1 correspondence']);
    try
        [rtheory, LI, Mapping] = ReorderSpheres (Spheres, theory, LI );
    catch
            disp ('failed to reorder spheres');
            rethrow(lasterror);
    end
    %save;
    %scatter;
    pdffile=sprintf('%s_aqual.pdf',ofile);
    [D,sfac,sunit]=ADNIScatterPlots(rtheory,Spheres,Headers,ofile,pdffile);
    rmstxt='';
    rmstxt=ADNIEvaluateLinearity(rtheory,Spheres,rmstxt);
    clear pdffile;
    % now make ascii output for database purposes
    % one element per line
    txtfile=sprintf('%s_aqual.txt',ofile);
    fid=fopen(txtfile,'w');
    fprintf(fid,'Aqual output for: %s\n',inDir);
    fprintf(fid,'this is Aqual version %g\n',Aqual_version);
    fprintf(fid,'Executed %s by user %s on host %s \n',datestr(now), ...
        getenv('USER'),getenv('HOSTNAME'));
    fprintf(fid, 'Contrast Properties\n');
    fprintf(fid, 'S,N,SNR:%f,%f,%f\n',S,N,SNR);
    Contrast=[Spheres.Spheres(Spheres.CNRSphereIndex).AverageIntensity];
    Contrast=sort(Contrast);
    fprintf(fid, 'ContrastSphereMeanIntensities:%f;%f;%f;%f\n',Contrast(1),Contrast(2),Contrast(3),Contrast(4));
    clear Contrast;
    fprintf(fid, 'Summary Geometry Properties (values are R/L, A/P, S/I)\n');
    fprintf(fid, 'Standard Deviation of Position Residuals (mm): %f,%f,%f\n',std(D));
    fprintf(fid, ['Estimated Scale factors to linearly correct gradient' ...
        ' (unitless):%f,%f,%f\n'],sfac);
    fprintf(fid, ['unit vector scale factors' ...
        ' (unitless):%f,%f,%f\n'],sunit);
    pfac=(sfac-1.0)*100.0;

    fprintf(fid,'Correction Values translated for Philips Console (X,Y,Z) %f,%f,%f\n',pfac(2),pfac(1),pfac(3));
    fprintf(fid,'Non-linear Evaluation:\n');
    fprintf(fid,'%s',rmstxt);
    fprintf(fid,'mean and sd of fiducial sphere average intensities:%f,%f\n',...
        mean([Spheres.Spheres(Spheres.FiducialSphereIndex(:)).AverageIntensity]),...
        std([Spheres.Spheres(Spheres.FiducialSphereIndex(:)).AverageIntensity]));


    fclose(fid);
    clear Image;
    try
        Headers.Image=[];
    catch
        disp('no struct Headers.Image -- Headers.Image not cleared');
    end
    Spheres.Header=[];
    clear LI;
    clear Mask;
    tmpofile=sprintf('%s_aqual.mat',ofile);
    %save(tmpofile);
catch
    %    save;
    txtfile=sprintf('%s_aqual.txt',ofile);
    fid=fopen(txtfile,'w');
    fprintf(fid,'ERROR IN PROCESSING');
    fclose(fid);
%    rethrow(lasterror);
%catch
    disp('UNKOWN FATAL ERROR');
end
end
