function [COM, VoxelPoints] = LocalizeSphere ( SubImage, Transform, Resolution, Sigma )
% COM = LocalizeSphere ( SubImage, Spacing )
% Find the center of mass of a sphere in the image, based on the
% Transform and a RANSAC parameter fit.
%  
  

% First get our contour points.
T = OtsuThreshold ( SubImage );


% Old method took the LOG on a slice by slice basis

% VoxelPoints = [];
% for zz = 1:size(SubImage,3)
%   [PixelCount, Area, Contour] = SliceArea ( SubImage(:,:,zz), T, 0, 5 );
%   if PixelCount
%     Contour = Contour';
%     tmp = ones(size(Contour,1),1) * zz;
%     VoxelPoints = [VoxelPoints; Contour tmp];
%   end
% end

% Try this another way, filter for the LOG 
% I = LaplacianOfGaussian ( SubImage, 1.2*Resolution(1), Resolution
% );
I = LaplacianOfGaussian ( SubImage, Sigma, Resolution );
VoxelPoints = [];
for zz = 1:size(I,3)
  [PixelCount, Area, Contour] = SliceArea ( SubImage(:,:,zz), T, 0, 5, I(:,:,zz) );
  if PixelCount
    Contour = Contour';
    tmp = ones(size(Contour,1),1) * zz;
    VoxelPoints = [VoxelPoints; Contour tmp];
  end
end


% Convert to RAS Coordinates
tmp = ones ( size ( VoxelPoints, 1 ), 1 );
RASPoints = Transform * [VoxelPoints tmp]';
RASPoints = RASPoints(1:3,:)';

WarningState = warning ( 'off' );

% M-Estimator
X = RASPoints;
% Initial guess
C = FitSphere ( X );
Epsilon = 10e-3;
Delta = 100;
Options = optimset ( 'MaxIter', 1, 'Display', 'off' );


while Delta > Epsilon
  d = SphereResiduals ( C, X );
  Sigma = sqrt ( 2.0 ) * median ( abs ( d ) );
  % Do one optimization step
  [NewC, foo, bar, foobar] = fminunc ( @SphereCostFunction, C, Options, X, Sigma );
  Delta = norm ( C - NewC );
  C = NewC;
end



% disp ( 'M-Estimator:' );
COM = C;
warning ( WarningState );
return

% Start our RANSAC
N = size(RASPoints,1);
% Allow each point to be at max .5mm off
Epsilon = .05;

% Do 500 iterations
M = 1000;
% Results is to store the solution, and inlier count
Results = zeros ( M, 5 );

for m = 1:M
  % Take a random 4 points
  II = randperm(N);
  c = FitSphere ( RASPoints(II(1:4),:) );
  d = abs ( SphereResiduals ( RASPoints, c ) );

  Results(m,1) = length ( find ( d < Epsilon ) );
  Results(m,2:5) = c;
end

% Now find the parameters of the best RANSAC
MatchIndex = find ( Results(:,1) == max( Results(:,1) ));
MatchIndex = MatchIndex(1);
c = Results(MatchIndex,2:5);
X = RASPoints;
d = abs ( SphereResiduals ( c, X ) );
InlierIndex = find ( d < Epsilon );
C = FitSphere ( RASPoints(InlierIndex,:) );
disp ( 'RANSAC;' );
C

function d = DistanceToSphere ( X, c )
% For each point in X(N,3), calculate the distance to the sphere
% identified by parameters c = [cx, cy, cz, r]
d = sqrt ( ( X(:,1) - c(1) ).^2 + ...
           ( X(:,2) - c(2) ).^2 + ...
           ( X(:,3) - c(3) ).^2 ) ...
    - r;
d = abs ( d );
    

function c = FitSphere ( X )
% X is Nx3
D = [X ones(size(X,1),1)];
A = D' * -1 * sum ( X'.^2 )';

T = inv(D' * D) * A;

c = -.5 * T(1:3);
r = sqrt ( -1 * T(4) + sum ( c.^2 ) );
c = [c' r];
return


function d = SphereResiduals ( c, X )
  d = sqrt ( ( X(:,1) - c(1) ).^2 + ...
             ( X(:,2) - c(2) ).^2 + ...
             ( X(:,3) - c(3) ).^2 ) ...
      - c(4);
  return

function d = SphereCostFunction ( c, X, Sigma )
  d = SphereResiduals ( c, X );
  d = d.^2;
  d = sum ( d ./ (Sigma.^2 + d ) );
  return
  
X = [ 1 0 0 1;
      0 1 0 1;
      0 0 1 1;
      -1 0 0 1];

CC = [1 -1 5]
X(:,1) = X(:,1) + CC(1);
X(:,2) = X(:,2) + CC(2);
X(:,3) = X(:,3) + CC(3);

Noise = .5 * ( rand ( 4,3 ) - .5 );

X(:,1:3) = X(:,1:3) + Noise;

D = X(:,1:3);
A = -1 * sum ( D'.^2 )'

T = inv(X) * A

c = -.5 * T(1:3)
r = -1 * T(4) + sum ( c.^2 )
