function [PhantomInformation, LabeledImage] = SegmentSpheres ( Header, Image, ExpectedSpheres )
% Properties = SegmentSpheres ( Header, Image, <Threshold>, <Neighbors> )
% Segment the image, finding the spheres.  Use the header
% information to calculate positions and center's of mass
% The relevant fields are:
%   Spheres -- List of found spheres
%   RASToCanonicalTransform -- 4x4 homogeneous transformation
%                                  from scanner coordinates to the
%                                  canonical coordinate system
%   VoxelToRASTransform -- transform from voxels to scanner
%                              coordinate systems
%   Spheres:
%     COM                -- Center of mass of the sphere in voxel coordinates
%     RASCOM             -- Center of mass of the sphere in RAS coordinates (mm)
%     CanonicalCOM       -- Center of mass in the phantom coordinate system (mm)
%     Voxels             -- count of voxels in this sphere
%     Volume             -- volume in mm^3 of this sphere
%     PCA                -- principal component vectors
%     PCAVariance        -- Variance in PCA vectors
%     AverageIntensity   -- Average Intensity in this sphere
%     CenterVoxel        -- Transpose of COM, used in WarpPhantom
%     CenterRAS          -- Transpose of RASCOM, used in WarpPhantom
%     Voxels             -- Number of voxels in the sphere
%     Volume             -- Volume of this sphere in mm^3
    
if ~exist ( 'ExpectedSpheres', 'var' )
  ExpectedSpheres = 171;
end

% These guys are the segmentation parameters
if ~exist ( 'ThresholdPercentage', 'var' )
  ThresholdPercentage = 0.3;  % Percentage of maximum intensity to
                              % call a sphere
end
if ~exist ( 'ConnectComponentsNeighbors', 'var' )
  ConnectComponentsNeighbors = 26;  % How many neighbors to use, can
                                    % be 6, 18, 26
end
                                  

UseWaitbar = 0;  

if ( isfield ( Header, 'mri' ) )
  resolution = [Header.mri.pixsize Header.mri.slice_thickness];
else
  resolution = [Header.PixelSpacing' Header.SpacingBetweenSlices];
end
voxelvolume = resolution(1) * resolution(2) * resolution(3);
Image = squeeze ( Image );
[x y z] = size ( Image );

Transform = CoordinateSystems ( Header );


% Threshold = ThresholdPercentage * max ( Image(:) );
Threshold = OtsuThreshold ( Image );
V = Image > Threshold;


% Connected components labeling of the spheres...
% Fall back to thresholding if we don't get 171 spheres!
[LabeledImage, SphereCount] = bwlabeln ( V, ConnectComponentsNeighbors );
if SphereCount < ExpectedSpheres
  Threshold = ThresholdPercentage * max ( Image(:) );
  V = Image > Threshold;
  [LabeledImage, SphereCount] = bwlabeln ( V, ConnectComponentsNeighbors );
end

% Calculate properties of these objects
Properties = struct([]);

PropertyList = { 'Area', 'Centroid', 'SubarrayIdx', ...
                 'FilledArea', 'PixelIdxList', 'PixelList' };


Stats = regionprops ( LabeledImage, PropertyList );
C = sum ( (voxelvolume*[Stats.Area]) >= 300.0 );
FoundSpheres = C == ExpectedSpheres;
ThresholdPercentage = 0.1;
ExponentialStepSize = 0.25;
LinearStepSize = 0.05;
NumberOfLinearSteps = 10;
StepSize = LinearStepSize;
StepSize = 0.1;

FewerSpheres = C < ExpectedSpheres;

ImageMax = max ( Image(:) );
TT = ThresholdPercentage * ImageMax;

while ~FoundSpheres
  
  % Figure out how many sheres we have
  Threshold = ThresholdPercentage * ImageMax;
  V = Image > Threshold;
  [LabeledImage, SphereCount] = bwlabeln ( V, ConnectComponentsNeighbors );
  Stats = regionprops ( LabeledImage, PropertyList );

  % Volume should be > 500mm^3
  NewC = sum ( (voxelvolume*[Stats.Area]) >= 300.0 );

  
  D = sign ( NewC - C );
  if D == 0
    if NewC == ExpectedSpheres
      FoundSpheres = 1;
    else
      D = 1;
    end
  end
  ThresholdPercentage = ThresholdPercentage + StepSize * D;
  disp ( sprintf ( '\tFound: %d Last Count: %d', NewC, C ) );
  C = NewC;
  FoundSpheres = C == ExpectedSpheres;
  disp ( sprintf ( '\tThreshold: %0.4f (%4d) found %d spheres', ThresholdPercentage, round(Threshold), C ) );
  
  SwitchDirection = FewerSpheres ~= ( C < ExpectedSpheres );

  if 0
  if NumberOfLinearSteps > 0 && ~SwitchDirection
    NumberOfLinearSteps = NumberOfLinearSteps - 1;
  else
    disp ( sprintf ( '\tSwitching Direction' ) );
    % Switch into exponential mode
    NumberOfLinearSteps = 0;
    StepSize = StepSize / 2.0;
    if StepSize < 0.0005
      FoundSpheres = 1;
    end
  end
  end
  % disp ( sprintf ( '\tSwitching Direction' ) );
  % Switch into exponential mode
  NumberOfLinearSteps = 0;
  StepSize = StepSize / 1.3;
  if StepSize < 0.0005
    FoundSpheres = 1;
  end
end

clear V




Stats = regionprops ( LabeledImage, PropertyList );
SphereID = 1;

for i = 1:SphereCount
  % disp ( sprintf ( '%d/%d', i, SphereCount ) );
  C = sum ( (voxelvolume*[Stats.Area]) >= 500.0 );
  
  if (voxelvolume*Stats(i).Area) < 500.0
    LabeledImage(find(LabeledImage==i)) = 0;
    % Silently ignore these spheres...
    warning ( sprintf ( '\tOnly found %d voxels in sphere, rejecting!!!!', Stats(i).Area ) );
    continue;
  end
  
  % Note!!!
  % regionprops returns locations in (x,y,z) coordinates, not
  % (i,j,k) as this code is expecting...  Need to exchange the x
  % and y columns...
  
  VoxelLocation = Stats(i).PixelList;
  VoxelLocation(:,[1 2]) = VoxelLocation(:,[2 1]);
  
  VoxelIntensity = Image(Stats(i).PixelIdxList);
  oo = ones ( length ( VoxelIntensity ), 1 );
  RASLocation = Transform * [VoxelLocation oo]';
  RASLocation = RASLocation(1:3,:)';
  
  Gray = sum ( VoxelIntensity );
  % RAS COM  
  XCOM = sum ( VoxelIntensity .* RASLocation(:,1) ) / Gray;
  YCOM = sum ( VoxelIntensity .* RASLocation(:,2) ) / Gray;
  ZCOM = sum ( VoxelIntensity .* RASLocation(:,3) ) / Gray;
  Properties(SphereID).RASCOM = [XCOM; YCOM; ZCOM];

  % Voxel COM
  XCOM = sum ( VoxelIntensity .* VoxelLocation(:,1) ) / Gray;
  YCOM = sum ( VoxelIntensity .* VoxelLocation(:,2) ) / Gray;
  ZCOM = sum ( VoxelIntensity .* VoxelLocation(:,3) ) / Gray;
  COM = [XCOM YCOM ZCOM];

  Properties(SphereID).COM = [XCOM; YCOM; ZCOM];
  % Properties(SphereID).VoxelCOM = [XCOM; YCOM; ZCOM];
  Properties(SphereID).AverageIntensity = Gray;

  % Get a subimage 6 pixels wider than the smallest bounding box
  % Determine the center of that...
  [SubImageIndex, Offset] = GetSubImageIndex ( Stats(i), 3, size(Image) );
  SubImage = Image(SubImageIndex{1}, SubImageIndex{2}, SubImageIndex{3} );
  Translation = eye(4);
  Translation(1:3,4) = (Offset-1)';
  
  if 1 == 0
    % Store where this came from
    Properties(SphereID).SubImageIndex = SubImageIndex;
    Properties(SphereID).Offset = Offset;
    
    [SphereModel, VoxelPoints] = LocalizeSphere ( SubImage, Transform * Translation, resolution, 1.0 );
    Properties(SphereID).ModelRASCOM = SphereModel(1:3)';
    Properties(SphereID).ModelRadius = SphereModel(4);
    
  
    Properties(SphereID).CenterRAS = Properties(SphereID).ModelRASCOM;
    Foo = inv ( Transform ) * [Properties(SphereID).CenterRAS; 1];
    Properties(SphereID).CenterVoxel = Foo(1:3)';
  
    d = norm ( Properties(SphereID).ModelRASCOM - Properties(SphereID).RASCOM );
    if d > 2.5
      % Properties(SphereID).CenterVoxel
      % Properties(SphereID).VoxelCOM'
      % VV = VolumeViewer ( SubImage, 'Red is Model, Blue is contour' );
      % VolumeViewer ( 'SetPoints', VV, Properties(SphereID).CenterVoxel - Offset );
      % VolumeViewer ( 'SetBluePoints', VV, VoxelPoints );
      warning ( sprintf ( '\tRAS Center of mass and model fit are different ID: %d norm == %f!', SphereID, d ) );
    end
  end
  % For the moment, stick with the RASCOM...
  Properties(SphereID).CenterRAS = Properties(SphereID).RASCOM;
  Foo = inv ( Transform ) * [Properties(SphereID).CenterRAS; 1];
  Properties(SphereID).CenterVoxel = Foo(1:3)';
  
  
  APrime = VoxelLocation - repmat ( COM, size ( VoxelLocation, 1 ), 1 );
  APrime = APrime' * APrime ./ size ( APrime, 1 );
  [V, D] = eig ( APrime );
  D = max ( D )';
  Properties(SphereID).Voxels = size ( VoxelLocation, 1 );  % Voxel count
  Properties(SphereID).Volume = Properties(SphereID).Voxels * voxelvolume;
  Properties(SphereID).PCA = D;                             % Principle components
  Properties(SphereID).PCAVariance = var ( D );
  
  LabeledImage(find(LabeledImage==i)) = SphereID;
  
  SphereID = SphereID + 1;
  
end

if length(Properties) ~= ExpectedSpheres
  error ( sprintf ( '\tFound %d spheres, not %d...', length(Properties), ExpectedSpheres ) );
end


PhantomInformation = struct([]);
[P T Idx SNRSphereIndex CNRSphereIndex FiducialSphereIndex] = CalculateCanonicalTransform ( Properties, ExpectedSpheres );
PhantomInformation(1).Spheres = P;
PhantomInformation.Threshold = Threshold;
PhantomInformation.RASToCanonicalTransform = T;
PhantomInformation.VoxelToRASTransform = Transform;
PhantomInformation.MatchError = 0.0;
PhantomInformation.LandmarkSphereIndex = Idx;
PhantomInformation.Exam = Header.StudyID;
PhantomInformation.Series = Header.SeriesNumber;
PhantomInformation.Header = Header;
PhantomInformation.SNRSphereIndex = SNRSphereIndex;
PhantomInformation.CNRSphereIndex = CNRSphereIndex;
PhantomInformation.FiducialSphereIndex = FiducialSphereIndex;

% disp ( sprintf ( 'Found: %d spheres returning %d', SphereCount, length(P) ) )

function [Properties, T, Idx, SNRSphereIndex, CNRSphereIndex, FiducialSphereIndex] = CalculateCanonicalTransform(Properties, ExpectedSpheres)
% CalculateCanonicalTransform - Determine canonical sphere numbers by transforming to regular space and relabeling each sphere
% Requires a Properties structure from ProcessPhantom
% Works with scanner COM, which is in mm and in the RAS coordinate system
  
  CenterRAS = [Properties.CenterRAS]';
  
  % Sort and take the three largest
  [Y, idx] = sort ( [Properties.Volume] );
  r = length(idx):-1:1;
  Y = Y(r);
  idx = idx(r);

  if ExpectedSpheres == 163
    % ADNI phantom, use 1, 4, & 5
    Sidx = idx([1 4 5]);
    
    % Cache the index of the SNR and CNR Spheres
    SNRSphereIndex = idx(1);
    CNRSphereIndex = idx([2 3]);
    FiducialSphereIndex = idx(4:end);
    
  else
    Sidx = idx(1:3);
  end
  S1 = Properties(Sidx(1)).CenterRAS;
  S2 = Properties(Sidx(2)).CenterRAS;
  S3 = Properties(Sidx(3)).CenterRAS;

  D = [];
  D(1) = sqrt ( sum ( ( S1 - S2 ).^2 ) );
  D(2) = sqrt ( sum ( ( S2 - S3 ).^2 ) );
  D(3) = sqrt ( sum ( ( S1 - S3 ).^2 ) );
  
  [DD, Didx] = sort ( D );
  
  % Find the sphere left out in the D calculations
  Map = [3 1 2];
  
  foo = Sidx(Map(Didx));
  Idx = foo;
  SZidx = foo(1);
  SXidx = foo(2);
  SOidx = foo(3);
  
  SO = CenterRAS(SOidx,:);
  SX = CenterRAS(SXidx,:);
  SZ = CenterRAS(SZidx,:);
  
  % Now we have the Origin, Zhat, Xhat spheres 
  % Define a coordinate system change that maps from the CenterRAS space,
  % to the internal coordinate system defined by the vectors.
  
  Ttranslate = eye(4);
  Ttranslate(1:3,4) = -SO';
  
  Trotate = eye(4);  
  Zhat = ( SZ - SO ) ./ norm ( SZ - SO );
  Xhat = ( SX - SO ) ./ norm ( SX - SO );
  Yhat = cross ( Zhat, Xhat );
  Xhat = cross ( Yhat, Zhat );
  
  Trotate(1:3,1:3) = [Xhat' Yhat' Zhat']';
  
  Tscale = eye(4) ./ norm ( SZ - SO );
  Tscale(4,4) = 1;
  
  % T is the transformation that transforms CenterRAS points to points in
  % "Phantom" space.
  % Don't want to scale.
  T = Trotate * Ttranslate;
  
  len = size ( CenterRAS, 1 );
  CenterRASPrime = [CenterRAS ones(len,1)]';
  NewCenterRAS = T * CenterRASPrime;
  NewCenterRAS = NewCenterRAS';
  NewCenterRAS = NewCenterRAS(:,1:3);
  
  for ii = 1:len,
    Properties(ii).CanonicalCOM = NewCenterRAS(ii,:)';
  end
  

  
