function [PhantomInformation, LabeledImage] = SegmentSpheresADNI ( Headers, Image, LabelMap, ExpectedSpheres )
% Properties = SegmentSpheres ( Headers, Image[, LabelMap][, ExpectedSpheres] )
% Segment the image, finding the spheres.  Use the header
% information to calculate positions and center's of mass
% If LabelMap exists and is not empty, use this as the sphere segmentation.
% The MayoPhantom is a flag to indicate the "leaky" Mayo phantom
% 
% In the ADNI phantom, there should be:
% 158 1cm spheres   (   523 mm^3)
% 2 1.5cm spheres   (  1767 mm^3)
% 4 3cm spheres     ( 14137 mm^3)
% 1 6cm sphere      (113097 mm^3)
%
% The relevant fields are:
%   Spheres -- List of found spheres
%   RASToCanonicalTransform -- 4x4 homogeneous transformation
%                                  from scanner coordinates to the
%                                  canonical coordinate system
%   VoxelToRASTransform -- transform from voxels to scanner
%                              coordinate systems
%   Spheres:
%     COM                -- Center of mass of the sphere in voxel coordinates
%     RASCOM             -- Center of mass of the sphere in RAS coordinates (mm)
%     CanonicalCOM       -- Center of mass in the phantom coordinate system (mm)
%     Voxels             -- count of voxels in this sphere
%     Volume             -- volume in mm^3 of this sphere
%     PCA                -- principal component vectors
%     PCAVariance        -- Variance in PCA vectors
%     AverageIntensity   -- Average Intensity in this sphere
%     CenterVoxel        -- Transpose of COM, used in WarpPhantom
%     CenterRAS          -- Transpose of RASCOM, used in WarpPhantom
%     Voxels             -- Number of voxels in the sphere
%     Volume             -- Volume of this sphere in mm^3

StartTime = clock;

MinimumVolume = 100;
% Calculate properties of these objects
Properties = struct([]);

PropertyList = { 'Area', 'Centroid', 'SubarrayIdx', ...
                 'FilledArea', 'PixelIdxList', 'PixelList' };

  
if ~exist ( 'ExpectedSpheres', 'var' )
  ExpectedSpheres = SpherePositions;
end

if ~exist ( 'MayoPhantom', 'var' )
  MayoPhantom = 0;
else
        MayoPhantom = 0;
            warning ('mayo phantom disabled');
end

if ~exist ( 'LabelMap', 'var' )
  LabelMap = [];
end

% These guys are the segmentation parameters
if ~exist ( 'ThresholdPercentage', 'var' )
  ThresholdPercentage = 0.3;  % Percentage of maximum intensity to
                              % call a sphere
end
if ~exist ( 'ConnectComponentsNeighbors', 'var' )
  ConnectComponentsNeighbors = 26;  % How many neighbors to use, can
                                    % be 6, 18, 26
end

UseWaitbar = 0;  

Header = Headers(1);
if ( isfield ( Header, 'mri' ) )
  resolution = [Header.mri.pixsize Header.mri.slice_thickness];
else
  if isfield ( Header, 'PixelSpacing' )
  % some MR vendors provide SpacingBetweenSlices, if so use that otherwise 
  % calculate an estimate from two adjacent slice positions
    if isfield (Header,'SpacingBetweenSlices')
    Z = Header.SpacingBetweenSlices;
    else
    Z = sqrt ( sum ( ( Headers(2).ImagePositionPatient - Headers(1).ImagePositionPatient ).^2) );
    end
  resolution = [Header.PixelSpacing' Z];
  else
   if isfield ( Header, 'NumVols' )
   % AVW structure
  if isfield ( Header, 'RelativeImageOrientationPatient' )
    a = SplitField ( Header.RelativeImageOrientationPatient, '\' );
    RowOrientation = a(1:3);
    ColOrientation = a(4:6);
  elseif isfield ( Header, 'Orientation' )
    switch Header.Orientation
     case 'Transverse'
      % Don't need to do anything
     case 'Coronal'
      RowOrientation = [1 0 0];
      ColOrientation = [0 0 1];
     case 'Sagittal'
      RowOrientation = [0 1 0];
      ColOrientation = [0 0 1];
    end
  end
  if isfield ( Header, 'VoxelWidth' )
 %     if isfield (Header, 'Orientation')
 %         if (Header.Orientation=='niftiRAS')
 %             resolution = [Header.VoxelHeight Header.VoxelWidth, Header.VoxelDepth];
 %         end
 %     else
          resolution = [Header.VoxelWidth Header.VoxelHeight, Header.VoxelDepth];
  %    end
  end
  end
  end
end
VoxelVolume = resolution(1) * resolution(2) * resolution(3);
Image = squeeze ( Image );
[x y z] = size ( Image );

Transform = CoordinateSystems ( Headers );

% Find the global threshold from the central CNS sphere
ImageMax = max ( Image(:) );
CurrentThreshold = ThresholdPercentage * ImageMax;

if isempty ( LabelMap )

  C = 10;
  
  [N,X] = hist ( Image(:), 300 );
  % Find the threshold where we get 80% of the spheres
  CC = VoxelVolume * cumsum ( N );
  
  VolumeCutoff = 158 * 523 + 2 * 1767 + 1 * 113097;
  VolumeCutoff = 0.8 * VolumeCutoff;
  
  Idx = find ( CC > (CC(end) - VolumeCutoff ) );
  
  CurrentThreshold = X(Idx(1));
  V = Image > CurrentThreshold;
  [LabeledImage, SphereCount] = bwlabeln ( V, ConnectComponentsNeighbors );
  Stats = regionprops ( LabeledImage, PropertyList );

  % Find the largest one
  [Y, idx] = sort ( VoxelVolume*[Stats.Area] );
  r = length(idx):-1:1;
  Y = Y(r);
  idx = idx(r);

  % Find a VOI around the SNR sphere that is 110% of the diameter
  SNRSphereVolume = Y(1);
  ExpandBy = round ( 0.1 * 2 * ( 3 / (4 * pi) * ExpectedSpheres.Spheres(ExpectedSpheres.SNRSphereIndex).Volume)^(1/3) );

  [SubImageIndex, Offset] = GetSubImageIndex ( Stats(idx(1)).SubarrayIdx, ExpandBy, size(Image) );
  SNR = Image(SubImageIndex{1}, SubImageIndex{2}, SubImageIndex{3} );

  CurrentThreshold = OtsuThreshold ( SNR );
  Count = 1;
  while 1
    % What is our current volume
    SNRTemp = SNR > CurrentThreshold;
    SNRTemp = bwlabeln ( SNRTemp, ConnectComponentsNeighbors );
    Center = round ( size(SNRTemp) / 2 );
    CenterIdx = sub2ind ( size(SNRTemp), Center(1), Center(2), Center(3) );
    SNRTemp = SNRTemp == SNRTemp(CenterIdx);
    SNRTemp = imfill ( SNRTemp, round ( size(SNR) / 2 ) );
    SNRSphereVolume = VoxelVolume * sum ( SNR(:) > CurrentThreshold );
    % disp ( sprintf ( '\tSNR Sphere Volume %6.1f @ Threshold: %d', SNRSphereVolume, round(CurrentThreshold) ) );
    % Upper spec limit is  120,000 mm^3
    if SNRSphereVolume > 120000.0
      % Raise the threshold
      CurrentThreshold = CurrentThreshold + 0.01*ImageMax;
    elseif SNRSphereVolume < 108000.0
      % Lower the threshold
      CurrentThreshold = CurrentThreshold - 0.01*ImageMax;
    else
      % We are in the sweet spot
      break
    end
    Count = Count + 1;
    if Count > 30
      warning 'Could not find central SNR sphere';
      break
    end
  end
  
  IdentifiedSpheres = zeros ( size ( Image ), 'int16' );
  IdentifiedSpheres(SubImageIndex{1}, SubImageIndex{2}, SubImageIndex{3} ) = SNRTemp;


  FiducialCount = 0;
  SphereDescriptions = {};
  % the following determines the search order for the spheres
  % at this point we should have found the gigantic SNR sphere
  % so next we look for the "pole" spheres and then
  % in what's leftover we pull out the little spheres
  % finally (and disabled 20050616) we could look for the CNR spheres
  % the lower contrast in the T1-darkest CNR sphere may or may not 
  % have ben found 
  % 1.5cm spheres  r=15mm/2
  SphereDescriptions{1}.Radius = 7.5;
  SphereDescriptions{1}.MinVolume = 1000.;
  SphereDescriptions{1}.MaxVolume = 2000;
  SphereDescriptions{1}.ExpectedSphereCount = 2;
  SphereDescriptions{1}.NCThresholdStart = 0.9;
  
  % 1cm spheres
  SphereDescriptions{2}.Radius = 5;
  SphereDescriptions{2}.MinVolume = 350.;
  SphereDescriptions{2}.MaxVolume = 650;
  SphereDescriptions{2}.ExpectedSphereCount = 158;
  SphereDescriptions{2}.NCThresholdStart = 0.8;

  % Not using 3cm spheres here
  % 3cm spheres
  SphereDescriptions{3}.Radius = 15;
  SphereDescriptions{3}.MinVolume = 12000.;
  SphereDescriptions{3}.MaxVolume = 16000.;
  SphereDescriptions{3}.ExpectedSphereCount = 4;
  SphereDescriptions{3}.NCThresholdStart = 0.8;
  
  % Here is the Mask to use, as spheres are found, don't recompute the
  % Normalized Correlation

  % We've found the SNR Sphere, do the NormalizedCorrelation3d
  % Skip the larger CNR spheres (DescriptionIdx=3)
  for DescriptionIdx = 1:3
    % Go through each of them, extract a local bounding box a little larger
    %than the currently sought spheres
    SphereRadius = SphereDescriptions{DescriptionIdx}.Radius;
    MinVolume = SphereDescriptions{DescriptionIdx}.MinVolume;
    MaxVolume = SphereDescriptions{DescriptionIdx}.MaxVolume;
    ExpectedSphereCount = SphereDescriptions{DescriptionIdx}.ExpectedSphereCount;
    BBSize = 3 + SphereRadius;
    % Make the sphere template of appropriate size
    [x,y,z] = meshgrid ( -BBSize:resolution(2):BBSize,  -BBSize:resolution(1):BBSize, -BBSize:resolution(3):BBSize );
    T = sqrt ( x.^2 + y.^2 + z.^2 ) < SphereRadius;
    
    ST = size ( T );
    A = PMNormalizedCorrelation3d ( Image, T(:,:,round(ST(3)/2)) );
    % A = NormalizedCorrelation3d ( Image, T, Mask );
      
    
    % Start at 0.7 as a threshold, then step down if we don't find
    % the spheres that we expect.
    foo = 1;
    NCThreshold = SphereDescriptions{DescriptionIdx}.NCThresholdStart;
    while 1
      % Find the regions
      [LabeledImage, SphereCount] = bwlabeln ( A > NCThreshold, ConnectComponentsNeighbors );
      % clear A;
      Stats = regionprops ( LabeledImage, PropertyList );
      PotentialPositions = zeros ( length ( Stats ), 3 );
      for ii = 1:length(Stats)
        PotentialPositions(ii,:) = Stats(ii).Centroid';
      end
      
      PPcount = size(PotentialPositions,1);
      Found = zeros ( PPcount, 1 );
      for ii = 1:PPcount
        SubIdx = {};
        Center = round ( PotentialPositions(ii,:) );
        Start = floor ( Center - SphereRadius ./ resolution ) - 1;
        End = floor ( Center + SphereRadius ./ resolution ) + 1;
%         for ttt = 1:3
%           if Start(ttt) < 1
%             Start(ttt) = 1;
%           end
%           if End(ttt) > size(Image,ttt)
%             End(ttt) = size(Image,ttt);
%           end
%           SubIdx{ttt} = [Start(ttt):End(ttt)];
%         end

if Start(1)<1
    Start(1)=1;
end
if End(1)>size(Image,2)
    End(1)=size(Image,2);
end
if Start(2)<1
    Start(2)=1;
end
if End(2)>size(Image,1)
    End(2)=size(Image,1);
end
if Start(3)<1
    Start(3)=1;
end
if End(3)>size(Image,3)
    End(3)=size(Image,3);
end
  for ttt = 1:3
        SubIdx{ttt} = [Start(ttt):End(ttt)];
         end



        % Note: centroid is x,y,z, == j,i,k not i,j,k as might be expected
        % Thus we need to switch SubIdx{2} and SubIdx{1}
        FiducialSphere = Image(SubIdx{2},SubIdx{1},SubIdx{3});
        FiducialThreshold = OtsuThreshold ( FiducialSphere );
        FiducialTemp = FiducialSphere > FiducialThreshold;
        % Keep only the biggest object in the threshold, to reject some of the junk
        FiducialLabel = bwlabeln ( FiducialTemp, ConnectComponentsNeighbors );
        FiducialStats = regionprops ( FiducialLabel, PropertyList );
        [m, idx] = max ( [FiducialStats.Area] );
        if isempty ( idx )
          continue
        end
        FiducialTemp = FiducialLabel == idx(1);
        
        
        % FiducialTemp = imfill ( FiducialTemp, round ( size(FiducialTemp) / 2 ) );
        FiducialVolume = sum ( FiducialTemp(:)) * VoxelVolume;
        % Stick candidates over of the right size in the final result
        if FiducialVolume > MinVolume && FiducialVolume < MaxVolume
          % disp ( sprintf ( 'Found a sphere(%d) at: %d, %d, %d Volume: %f', ii, Center(1), Center(2), Center(3), FiducialVolume ) );
          FiducialCount = FiducialCount + 1;
          IdentifiedSpheres(SubIdx{2},SubIdx{1},SubIdx{3}) = FiducialTemp | IdentifiedSpheres(SubIdx{2},SubIdx{1},SubIdx{3});
          
          % No need to reprocess these voxels
          Found(ii) = 1;
        else
          % disp ( sprintf ( 'Rejected a sphere(%d) at: %d, %d, %d Volume: %f', ii, Center(1), Center(2), Center(3), FiducialVolume ) );
        end
      end

      if NCThreshold < 0.5
        warning ( sprintf ( 'Normalized correlation threshold is very low, continuing' ) );
        break;
      end
      
      if sum ( Found ) >= ExpectedSphereCount
        break;
      end
      if sum ( Found ) ~= ExpectedSphereCount
        disp ( sprintf ( 'Expected to find %d spheres at this stage, found %d instead', ExpectedSphereCount, sum ( Found ) ) );
      end
      if sum ( Found ) < ExpectedSphereCount
        NCThreshold = NCThreshold - 0.05;
        disp ( sprintf ( 'Decreasing NCThreshold to %0.2f', NCThreshold ) );
      end
      
    end
  end
else
  % disp ( 'Using previously segmented results' );
  IdentifiedSpheres = LabelMap;
end; % if we don't have a segmentation.
  

% Now, all candidate spheres should be in the IndentifiedSpheres array...
% there may be more than just the spheres however 
[LabeledImage, SphereCount] = bwlabeln ( IdentifiedSpheres, ConnectComponentsNeighbors );
Stats = regionprops ( LabeledImage, PropertyList );
C = sum ( (VoxelVolume*[Stats.FilledArea]) >= MinimumVolume );
% disp ( sprintf ( '\tFound %d spheres total', C ) );
[Y, idx] = sort ( VoxelVolume*[Stats.FilledArea] );
r = length(idx):-1:1;
Y = Y(r);
idx = idx(r);

% Identify the SNRSphere
SNRSphereId = idx(1);
Poles = [];
CNRSpheres = [];
Fiducials = [];

% Center is clearly the first one, mask out any spheres
% where the volume is 2000 < VolumeOfSphere < 15000
% Takes the first 158 Fiducials, rejects others
for ii = idx
  % Already handled
  if ii == SNRSphereId
    continue
  end
  VolumeOfSphere = VoxelVolume * Stats(ii).FilledArea;
  % disp ( sprintf ( '\tVolume: %4.f', VolumeOfSphere ) );
  if VolumeOfSphere > 2000.0 
    % disp ( sprintf ( '\tFound Contrast Sphere: %d Volume: %4.f', ii, VolumeOfSphere ) );
    CNRSpheres = [CNRSpheres ii];
    % LabeledImage(Stats(ii).PixelIdxList) = 0;
  end
  if VolumeOfSphere > 700.0 && VolumeOfSphere <= 2000.0
    % This is one of the poles
    Poles = [Poles ii];
  elseif VolumeOfSphere <= 700.0
    % This is a fiducial
    Fiducials = [Fiducials ii];
  end
end

if length ( Poles ) ~= 2
  warning 'Could not find the "pole" spheres';
end

if length ( Fiducials ) ~= 158
  % warning ( sprintf ( 'Found %d fiducials, should be 158', length ( Fiducials ) ) );
end

% Everything should be properly masked out.
SphereID = 1;
OK = zeros ( length(Stats), 1 );
OK(Poles) = 1;
OK(Fiducials) = 1;
OK(CNRSpheres) = 1;
OK(SNRSphereId) = 1;

Intensity = zeros(1,length(Stats) );
for i = 1:length(Stats)
  TT = mean ( Image(Stats(i).PixelIdxList) );
  Intensity(i) = TT;
end

MeanIntensity = mean ( Intensity );
StdIntensity = std ( Intensity );

if length ( Stats ) < 10
  error ( 'Found less than 10 Spheres!' );
end

for i = 1:length(Stats)
  % Skip small spheres
  if length(Stats(i).PixelIdxList) * VoxelVolume < MinimumVolume
    LabeledImage(find(LabeledImage==i)) = 0;
    continue;
  end
end

[LabeledImage, SphereCount] = bwlabeln ( LabeledImage>0, ConnectComponentsNeighbors );
Stats = regionprops ( LabeledImage, PropertyList );

% Reidentify the SNR sphere
[Y, idx] = sort ( VoxelVolume*[Stats.FilledArea] );
r = length(idx):-1:1;
Y = Y(r);
idx = idx(r);

% Identify the SNRSphere
SNRSphereId = idx(1);


for i = 1:length(Stats)
  % disp ( sprintf ( '%d/%d', i, SphereCount ) );
 
  if ~OK(i) 
    % LabeledImage(find(LabeledImage==i)) = 0;
    % continue;
  end
  if Stats(i).Area == 0
    LabeledImage(find(LabeledImage==i)) = 0;
    continue;
  end
  
  % Skip small spheres
  if length(Stats(i).PixelIdxList) * VoxelVolume < MinimumVolume
    LabeledImage(find(LabeledImage==i)) = 0;
    continue;
  end
  
  
  % Note!!!
  % regionprops returns locations in (x,y,z) coordinates, not
  % (i,j,k) as this code is expecting...  Need to exchange the x
  % and y columns...
  
  VoxelLocation = Stats(i).PixelList;
  VoxelLocation(:,[1 2]) = VoxelLocation(:,[2 1]);
  
  VoxelIntensity = Image(Stats(i).PixelIdxList);
  oo = ones ( length ( VoxelIntensity ), 1 );
  RASLocation = Transform * [VoxelLocation oo]';
  RASLocation = RASLocation(1:3,:)';
  
  Gray = sum ( VoxelIntensity );
  % RAS COM  
  XCOM = sum ( VoxelIntensity .* RASLocation(:,1) ) / Gray;
  YCOM = sum ( VoxelIntensity .* RASLocation(:,2) ) / Gray;
  ZCOM = sum ( VoxelIntensity .* RASLocation(:,3) ) / Gray;
  Properties(SphereID).RASCOM = [XCOM; YCOM; ZCOM];

  % Voxel COM
  XCOM = sum ( VoxelIntensity .* VoxelLocation(:,1) ) / Gray;
  YCOM = sum ( VoxelIntensity .* VoxelLocation(:,2) ) / Gray;
  ZCOM = sum ( VoxelIntensity .* VoxelLocation(:,3) ) / Gray;
  COM = [XCOM YCOM ZCOM];

  Properties(SphereID).COM = [XCOM; YCOM; ZCOM];
  % Properties(SphereID).VoxelCOM = [XCOM; YCOM; ZCOM];
  Properties(SphereID).AverageIntensity = Gray / length(VoxelIntensity);

  % Get a subimage 6 pixels wider than the smallest bounding box
  % Determine the center of that...
  [SubImageIndex, Offset] = GetSubImageIndex ( Stats(i).SubarrayIdx, 3, size(Image) );
  SubImage = Image(SubImageIndex{1}, SubImageIndex{2}, SubImageIndex{3} );
  Translation = eye(4);
  Translation(1:3,4) = (Offset-1)';
  
  Properties(SphereID).SubImageIndex = SubImageIndex;
  % For the moment, stick with the RASCOM...
  Properties(SphereID).CenterRAS = Properties(SphereID).RASCOM;
  Foo = inv ( Transform ) * [Properties(SphereID).CenterRAS; 1];
  Properties(SphereID).CenterVoxel = Foo(1:3)';
  
  
  APrime = VoxelLocation - repmat ( COM, size ( VoxelLocation, 1 ), 1 );
  APrime = APrime' * APrime ./ size ( APrime, 1 );
  [V, D] = eig ( APrime );
  D = max ( D )';
  Properties(SphereID).Voxels = size ( VoxelLocation, 1 );  % Voxel count
  Properties(SphereID).Volume = Properties(SphereID).Voxels * VoxelVolume;
  Properties(SphereID).PCA = D;                             % Principle components
  Properties(SphereID).PCAVariance = var ( D );

  % Do a sanity check here
  % If the intensity is less than 2 standard deviations below, drop
  % it
  if Intensity(i) < (MeanIntensity - 2.5*StdIntensity)
    % disp ( sprintf ( 'Found low intensity sphere @ %d', i ) );
    % LabeledImage(find(LabeledImage==i)) = 0;
    % continue;
  end
  

  
  LabeledImage(find(LabeledImage==i)) = SphereID;
  Poles(find(Poles==i)) = SphereID;
  SNRSphereId(find(SNRSphereId==i)) = SphereID;
  Fiducials(find(Fiducials==i)) = SphereID;
  CNRSpheres(find(CNRSpheres==i)) = SphereID;
  
  
  SphereID = SphereID + 1;
  
end


PhantomInformation = struct([]);
% Poles
[P T Idx] = CalculateCanonicalTransformADNI ( Properties );

% Because the Mayo phantom is leaking, figure out if we have
% captured the leak
if MayoPhantom
  % The position is about 1.5, 47, 0 in Canonical space.
  % See if any of the spheres come close, and reject
  Locations = [P.CanonicalCOM];
  BadLocation = [1.5 47 0]';
  Diff = Locations - repmat ( BadLocation, 1, size(Locations,2) );
  Diff = sqrt ( sum ( Diff.^2 ) )';
  
  if ( min ( Diff ) < 13. )
    % Reject this sphere
    BadIdx = find ( Diff == min(Diff) );
    disp ( sprintf ( 'Found bad sphere at: %d', BadIdx ) );
    LabeledImage(find(LabeledImage==BadIdx)) = 0;
    for idx = BadIdx+1:length(P)
      P(idx-1) = P(idx);
      LabeledImage(find(LabeledImage==idx)) = idx-1;
      Poles(find(Poles==idx)) = idx-1;
      Idx(find(Idx==idx)) = idx-1;
      SNRSphereId(find(SNRSphereId==idx)) = idx-1;
    end
    P = P(1:length(P)-1);
    disp ( sprintf ( 'New length: %d', length(P) ) );
    disp ( sprintf ( 'Max object: %d', max(LabeledImage(:)) ) );
  end
end

PhantomInformation(1).Spheres = P;
PhantomInformation.Threshold = CurrentThreshold;
PhantomInformation.RASToCanonicalTransform = T;
PhantomInformation.VoxelToRASTransform = Transform;
PhantomInformation.MatchError = 0.0;
PhantomInformation.LandmarkSphereIndex = Idx;
if isfield ( Header, 'StudyID' )
  PhantomInformation.Exam = Header.StudyID; 
else
  PhantomInformation.Exam = 1;
end
if isfield (Header,'SeriesNumber')
  PhantomInformation.Series = Header.SeriesNumber; 
else
  PhantomInformation.Series = 1;
end
PhantomInformation.Header = Header;
PhantomInformation.SNRSphereIndex = SNRSphereId;
PhantomInformation.CNRSphereIndex = CNRSpheres;
PhantomInformation.FiducialSphereIndex = Fiducials;

ProcessingTime = etime ( clock, StartTime );
disp ( sprintf ( 'Found: %d spheres in %0.4f seconds', length(PhantomInformation.Spheres), ProcessingTime ) )


