function W = TPS ( Source, Landmark, BasisType, varargin )
% W = TPS ( Source, Landmark, BasisType[, options] )
% Creates a structure that will warp a point in Source space into
% Landmark space.  
% BasisType is one of: 'vs', 'tps', 'ebs', 'ebrs'
%   'vs'   - Volume Spline           G(x) = r(x)^3 I
%   'tps'  - Thin Plate Spline       G(x) = r(x) I
%   'ebs'  - Elastic Body Spline     G(x) = ( a r(x)^2 I - 3 x x' ) r(x)
%   'ebrs' - Elastic Body Reciprocal Spline G(x) = ( b r(x) I - x x' / r(x)
% Optional arguments:
%   'Alpha', number : sets the alpha parameter, corresponding to
%   the EBS stiffness (Alpha) and EBRS stiffness (Beta)
%   'Stiffness', number : degree of interpolation 0 ->
%   interpolating, ~0.001 -> approximating, > 0.1 -> nearly affine
% Notes:
%   For the 'ebs', Alpha = 12 ( 1 - nu ) - 1
%   For the 'ebrs', Beta = 8 ( 1 - nu ) - 1
% See TPSValue.
  [D, N] = size ( Source );
  
  if D ~= 3
    error ( 'TPS requires 3 dimensional data' );
  end
  
  % Basis types, R2LogR2 = 1, R = 2
  % EBS = 3
  VSConstant = 1;
  TPSConstant = 2;
  EBSConstant = 3;
  EBRSConstant = 4;

  % Determine valid BasisType
  if ischar ( BasisType )
    switch lower ( BasisType )
     case 'vs'
      BasisType = VSConstant;
     case 'tps'
      BasisType = TPSConstant;
     case 'ebs'
      BasisType = EBSConstant;
     case 'ebrs'
      BasisType = EBRSConstant;
     otherwise
      error ( sprintf ( 'Unknown spline type: %s, must be one of ''vs'', ''tps'', ''ebs'' ''ebrs''', BasisType ) );
    end
  else
    if BasisType < VSConstant | BasisType > EBSConstant
      error ( sprintf ( 'Unknown spline type: %d, must be one of ''vs'', ''tps'', ''ebs'', ''ebrs''', BasisType ) );
    end
  end
  
  % Parse optional arguments
  Alpha = 0.25;
  Stiffness = 0.0;
  for ii = 1:2:length(varargin)
    Name = varargin{ii};
    Value = varargin{ii+1};
    switch lower ( Name )
     case 'stiffness',
      Stiffness = Value;
     case 'alpha',
      Alpha = Value;
     otherwise
      error ( sprintf ( 'Unknown option %s', Name ) );
    end
  end

  K = zeros ( 3*N, 3*N );
  ITmp = eye ( D, D );
  for row = 1:N
    for col = row:N
      if row ~= col
        x = Source(:,row) - Source(:,col);
        r = norm ( x );
        % Form G
        G = zeros ( D, D );
        switch BasisType
         case VSConstant
          G = ITmp * r.^3;
         case TPSConstant
          G = ITmp * r;
         case EBSConstant
          G = r * ( Alpha * r.^2 * ITmp - 3 * x * x' );
         case EBRSConstant
          if r > 1e-8
            rinv = 1.0 / r;
          else
            rinv = 0.0;
          end
          G = ( Alpha * r * ITmp -  rinv * x * x' );
	end
      else
        % Stiffness comes into play here
        G = Stiffness * ITmp;
      end
      % Put G into the proper place in K
      irow = 3*(row-1) + 1; icol = 3*(col-1) + 1;
      K(irow:irow+2, icol:icol+2) = G;
      K(icol:icol+2, irow:irow+2) = G;
    end
  end
  
  P = kron ( [Source' ones(N,1)], eye(3) );
  
  s = 3*N + 12;
  L = zeros ( s, s );
  L(1:3*N,1:3*N) = K;
  
  L(1:3*N,3*N+1:s) = P;
  L(3*N+1:s,1:3*N) = P';
  
  Y = [(Landmark(:) - Source(:))' zeros(1,12)]';
  
  % This is the W matrix from the Davis paper
  % Weights = inv ( L ) * Y;
  % Rescale this matrix for more accuracy.
  % Multiply both sides of the equation by a diagonal matrix
  % composed of 1/norm of each row of L.
  nn = sqrt ( sum ( (L.^2)' ) );
  NL = diag ( 1./nn, 0 );
  Weights = inv ( NL * L ) * NL * Y;
  % Weights = inv ( L ) * Y;
  
  W = struct([]);
  W(1).P = 1; % Initial creation
  W.P = Source';
  
  % Need to reshape the W matrix
  WTemp = Weights(1:3*N);
  
  W.Weights = reshape ( WTemp, D, N );
  
  CTemp = Weights(3*N+1:end);
  
  W.Constants = reshape ( CTemp, D, D+1 );
  W.ARotation = W.Constants(:,1:D);
  W.ATranslation = W.Constants(:,D+1);
  W.N = N;
  W.D = D;
  % W.Constants = Weights(N+1:end,:);
  switch BasisType
   case VSConstant
    W.BasisFunction = 'vs';
   case TPSConstant
    W.BasisFunction = 'tps';
   case EBSConstant
    W.BasisFunction = 'ebs';
   case EBRSConstant
    W.BasisFunction = 'ebrs';
   otherwise
    error ( sprintf ( 'Unknown basis type %d', BasisType ) );
  end
  W.BasisFunctionNumber = BasisType;
  W.Alpha = Alpha;
  W.Stiffness = Stiffness;
