#include "mex.h"

/* This is very important! Otherwise log links incorrectly... */
#include <math.h>
/*
 * mTPSValue.mex -- do the core TSP calculations
 *
 */

/* The gateway routine */
/* Arguments are: 
 * P - basis points (3xNumberOfBasisPoints)
 * W - basis weights (3xNumberOfBasisPoints)
 * X - list of points (3xNumberOfPoints)
 */

#define VS 1
#define TPS 2
#define EBS 3
#define EBRS 4


void VSFunction ( double r, double x[3], double G[3][3], double Alpha )
{
  G[0][0] = G[1][1] = G[2][2] = r * r * r;
}

void TPSFunction ( double r, double x[3], double G[3][3], double Alpha )
{
  G[0][0] = G[1][1] = G[2][2] = r;
}

void EBSFunction ( double r, double x[3], double G[3][3], double Alpha )
{
  double factor, radial, value, xi;
  int i, j;
  factor = -3.0 * r;
  radial = Alpha * r * r * r;
  for ( i = 0; i < 3; i++ )
    {
    xi = x[i] * factor;
    /* G is symmetric */
    for ( j = 0; j < i; j++ )
      {
      value = xi * x[j]; 
      G[i][j] = value;
      G[j][i] = value;
      }
    G[i][i] =  radial + xi * x[i];
    }
}

void EBRSFunction ( double r, double x[3], double G[3][3], double Alpha )
{
  double factor, radial, value, xi;
  int i, j;
  factor = ( r > 1e-8 ) ? ( -1.0 / r ) : 0.0;
  radial = Alpha * r;
  for ( i = 0; i < 3; i++ )
    {
    xi = x[i] * factor;
    /* G is symmetric */
    for ( j = 0; j < i; j++ )
      {
      value = xi * x[j]; 
      G[i][j] = value;
      G[j][i] = value;
      }
    G[i][i] =  radial + xi * x[i];
    }
}


void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[])
{
  double *P, *X, *W, *PointsPointer, *WeightsPointer;
  double *Out;
  int    status, NumberOfBasisPoints, NumberOfPoints, i;
  int D;
  int BasisIndex, PointsIndex, BasisType;
  char Buffer[1024];

  void (*BasisFunctionPointer) ( double r, double x[3], double G[3][3], double Alpha );
  
  double Alpha = 0.25;

  /*  Check for proper number of arguments. */
  /* NOTE: You do not need an else statement when using
     mexErrMsgTxt within an if statement. It will never
     get to the else statement if mexErrMsgTxt is executed.
     (mexErrMsgTxt breaks you out of the MEX-file.) 
  */
  if(nrhs < 3)
    {
    mexErrMsgTxt("Three inputs required.");
    }
  if(nlhs != 1)
    {
    mexErrMsgTxt("One output required.");
    }
  /*
  for ( i = 0; i < nrhs; i++ )
    {
    sprintf ( Buffer, "Input %d is %dx%d (MxN)", i, mxGetM(prhs[i]), mxGetN(prhs[i]) );
    mexWarnMsgTxt ( Buffer );
    }
  */

  BasisType = TPS;
  if ( nrhs >= 4 )
    {
    /* Get the scalar value */
    if ( !mxIsDouble(prhs[3]) || mxIsComplex(prhs[3])
         || !( mxGetM(prhs[3]) == 1 && mxGetN(prhs[3]) == 1 ) )
      {
      mexErrMsgTxt ( "Basis type must be a noncomplex scalar double" );
      }
    else
      {
      BasisType = mxGetScalar ( prhs[3] );
      }
    switch ( BasisType )
      {
      case VS:
        BasisFunctionPointer = VSFunction;
        break;
      case TPS:
        BasisFunctionPointer = TPSFunction;
        break;
      case EBS:
        BasisFunctionPointer = EBSFunction;
        break;
      case EBRS:
        BasisFunctionPointer = EBRSFunction;
        break;
      default:
        BasisType = TPS;
        BasisFunctionPointer = TPSFunction;
        break;
      }
    }

  if ( nrhs >= 5 )
    {
    /* Get the scalar value */
    if ( !mxIsDouble(prhs[4]) || mxIsComplex(prhs[4])
         || !( mxGetM(prhs[4]) == 1 && mxGetN(prhs[4]) == 1 ) )
      {
      mexErrMsgTxt ( "Alpha must be a noncomplex scalar double" );
      }
    else
      {
      Alpha = mxGetScalar ( prhs[4] );
      }
    }
  D = mxGetM ( prhs[0] );
  if ( mxGetM ( prhs[0] ) != D
       || mxGetM ( prhs[1] ) != D 
       || mxGetM ( prhs[2] ) != D )
    {
    mexErrMsgTxt("Inputs must be DxN.");
    }
  if ( D != 2 && D != 3 )
    {
    mexErrMsgTxt ( "Can only handle 2 or 3 dimensional data" );
    }
  if ( mxGetN ( prhs[0] ) != mxGetN ( prhs[1] ) )
    {
    mexErrMsgTxt ( "BasisPoints and Weights must have the same number of rows!" );
    }

  /* Get Pointers */
  P = mxGetPr ( prhs[0] );
  W = mxGetPr ( prhs[1] );
  X = mxGetPr ( prhs[2] );

  /* Sizes */
  NumberOfBasisPoints = mxGetN ( prhs[0] );
  NumberOfPoints = mxGetN ( prhs[2] );
  
  /* Set the output pointer to the output matrix. */
  plhs[0] = mxCreateDoubleMatrix ( D, NumberOfPoints, mxREAL );
  Out = mxGetPr ( plhs[0] );

  PointsPointer = X;
  for ( PointsIndex = 0; PointsIndex < NumberOfPoints; PointsIndex++ )
    {
    /* for each point, calculate U */
    double R2, t, G[3][3], x[3], r;
    double *BasisPointer;
    int i, j;
    BasisPointer = P;
    WeightsPointer = W;

    /* Zero out output */
    Out[0] = Out[1] = Out[2] = 0.0;

    /*
    sprintf ( Buffer, "Point: %d %f,%f,%f", PointsIndex, PointsPointer[0], PointsPointer[1], PointsPointer[2] );
    mexWarnMsgTxt ( Buffer );
    */

    for ( BasisIndex = 0; BasisIndex < NumberOfBasisPoints; BasisIndex++ )
      {
      /* Subtract */
      x[0] = PointsPointer[0] - BasisPointer[0];
      x[1] = PointsPointer[1] - BasisPointer[1];
      x[2] = PointsPointer[2] - BasisPointer[2];
      r = sqrt ( x[0] * x[0] + x[1] * x[1] + x[2] * x[2] );

      for ( i = 0; i < 3; i++ ) 
        {
        G[i][0] = 0.0;
        G[i][1] = 0.0;
        G[i][2] = 0.0;
        }

      (BasisFunctionPointer) ( r, x, G, Alpha );
#if 0
      switch ( BasisType )
        {
        case VS:
          {
          /* G[0][0] = G[1][1] = G[2][2] = r * r * r; */
          VSFunction ( r, x, G, Alpha );
          break;
          }
        case TPS:
          {
          /* G[0][0] = G[1][1] = G[2][2] = r; */
          TPSFunction ( r, x, G, Alpha );
          break;
          }
        case EBS:
          {
          EBSFunction ( r, x, G, Alpha );
          break;
          }
        case EBRS:
          {
          EBRSFunction ( r, x, G, Alpha );
          break;
          }
        }
#endif     
      for ( i = 0; i < 3; i++ )
        {
        for ( j = 0; j < 3; j++ )
          {
          Out[i] += G[i][j] * WeightsPointer[j];
          }
        }
      WeightsPointer += D;
      BasisPointer += D;
      }
    Out += D;
    PointsPointer += D;
    }

}
